"""
stp protocl support class(es)

http://libvirt.org/formatnwfilter.html#nwfelemsRulesProtoSTP
"""

from virttest.libvirt_xml import accessors, xcepts
from virttest.libvirt_xml.nwfilter_protocols import base


class Stp(base.TypedDeviceBase):

    """
    Create new Stp xml instances

    Properties:
        attrs: libvirt_xml.nwfilter_protocols.Stp.Attr instance
    """

    __slots__ = ('attrs',)

    def __init__(self, type_name='file', virsh_instance=base.base.virsh):
        accessors.XMLElementNest('attrs', self, parent_xpath='/',
                                 tag_name='stp', subclass=self.Attr,
                                 subclass_dargs={
                                     'virsh_instance': virsh_instance})
        super(Stp, self).__init__(protocol_tag='stp', type_name=type_name,
                                  virsh_instance=virsh_instance)

    def new_attr(self, **dargs):
        """
        Return a new Attr instance and set properties from dargs

        :param dargs: dict of attributes
        :return: new Attr instance
        """
        new_one = self.Attr(virsh_instance=self.virsh)
        for key, value in list(dargs.items()):
            setattr(new_one, key, value)
        return new_one

    def get_attr(self):
        """
        Return stp attribute dict

        :return: None if no stp in xml, dict of stp's attributes.
        """
        try:
            stp_node = self.xmltreefile.reroot('/stp')
        except KeyError as detail:
            raise xcepts.LibvirtXMLError(detail)
            node = stp_node.getroot()
            stp_attr = dict(list(node.items()))

        return stp_attr

    class Attr(base.base.LibvirtXMLBase):

        """
        Stp attribute XML class

        Properties:

        srcmacaddr: string, MAC address of sender
        srcmacmask: string, Mask applied to MAC address of sender
        type: string, Bridge Protocol Data Unit (BPDU) type
        flags: string, BPDU flag
        root_priority: string, Root priority (range start)
        root_priority_hi: string, Root priority range end
        root_address: string, Root MAC address
        root_address_mask: string, BPDU sender MAC address
        root_cost: string, Root path cost (range start)
        root_cost_hi: string, Root path cost range end
        sender_priority: string, Sender priority (range start)
        sender_priority_hi: string, Sender priority range end
        sender_address: string, BPDU sender MAC address
        sender_address_mask: string, BPDU sender MAC address mask
        port: string, Port identifier (range start)
        port_hi: string, Port identifier range end
        msg_age: string, Message age timer (range start)
        msg_age_hi: string, Message age timer range end
        max_age: string, Maximum age timer (range start)
        max_age_hi: string, Maximum age timer range end
        hello_time: string, Hello time timer (range start)
        hello_time_hi: string, Hello time timer range end
        forward_delay: string, Forward delay (range start)
        forward_delay_hi: string, Forward delay range end
        comment: string, text with max. 256 characters
        """

        __slots__ = ('srcmacaddr', 'srcmacmask', 'type', 'flags',
                     'root_priority', 'root_priority_hi', 'root_address',
                     'root_address_mask', 'root_cost', 'root_cost_hi',
                     'sender_priority', 'sender_priority_hi', 'sender_address',
                     'sender_address_mask', 'port', 'port_hi', 'msg_age',
                     'msg_age_hi', 'max_age', 'max_age_hi', 'hello_time',
                     'hello_time_hi', 'forward_delay', 'forward_delay_hi',
                     'comment')

        def __init__(self, virsh_instance=base.base.virsh):
            accessors.XMLAttribute('srcmacaddr', self, parent_xpath='/',
                                   tag_name='stp', attribute='srcmacaddr')
            accessors.XMLAttribute('srcmacmask', self, parent_xpath='/',
                                   tag_name='stp', attribute='srcmacmask')
            accessors.XMLAttribute('type', self, parent_xpath='/',
                                   tag_name='stp', attribute='type')
            accessors.XMLAttribute('flags', self, parent_xpath='/',
                                   tag_name='stp', attribute='flags')
            accessors.XMLAttribute('root_priority', self, parent_xpath='/',
                                   tag_name='stp', attribute='root-priority')
            accessors.XMLAttribute('root_priority-hi', self, parent_xpath='/',
                                   tag_name='stp', attribute='root-priority-hi')
            accessors.XMLAttribute('root_address', self, parent_xpath='/',
                                   tag_name='stp', attribute='root-address')
            accessors.XMLAttribute('root_address_mask', self,
                                   parent_xpath='/', tag_name='stp',
                                   attribute='root-address-mask')
            accessors.XMLAttribute('root_cost', self, parent_xpath='/',
                                   tag_name='stp', attribute='root-cost')
            accessors.XMLAttribute('root_cost_hi', self, parent_xpath='/',
                                   tag_name='stp', attribute='root-cost-hi')
            accessors.XMLAttribute('sender_priority', self, parent_xpath='/',
                                   tag_name='stp', attribute='sender-priority')
            accessors.XMLAttribute('sender_priority_hi', self,
                                   parent_xpath='/', tag_name='stp',
                                   attribute='sender-priority-hi')
            accessors.XMLAttribute('sender_address', self, parent_xpath='/',
                                   tag_name='stp', attribute='sender-address')
            accessors.XMLAttribute('sender_address_mask', self,
                                   parent_xpath='/', tag_name='stp',
                                   attribute='sender-address-mask')
            accessors.XMLAttribute('port', self, parent_xpath='/',
                                   tag_name='stp', attribute='port')
            accessors.XMLAttribute('port_hi', self, parent_xpath='/',
                                   tag_name='stp', attribute='port_hi')
            accessors.XMLAttribute('msg_age', self, parent_xpath='/',
                                   tag_name='stp', attribute='msg-age')
            accessors.XMLAttribute('msg_age_hi', self, parent_xpath='/',
                                   tag_name='stp', attribute='msg-age-hi')
            accessors.XMLAttribute('max-age', self, parent_xpath='/',
                                   tag_name='stp', attribute='max-age')
            accessors.XMLAttribute('max-age-hi', self, parent_xpath='/',
                                   tag_name='stp', attribute='max-age-hi')
            accessors.XMLAttribute('hello-time', self, parent_xpath='/',
                                   tag_name='stp', attribute='hello-time')
            accessors.XMLAttribute('hello-time-hi', self, parent_xpath='/',
                                   tag_name='stp', attribute='hello-time-hi')
            accessors.XMLAttribute('forward-delay', self, parent_xpath='/',
                                   tag_name='stp', attribute='forward-delay')
            accessors.XMLAttribute('forward-delay-hi', self, parent_xpath='/',
                                   tag_name='stp', attribute='forward-delay-hi')
            accessors.XMLAttribute('comment', self, parent_xpath='/',
                                   tag_name='stp', attribute='comment')

            super(self.__class__, self).__init__(virsh_instance=virsh_instance)
            self.xml = '<stp/>'



from distutils import dir_util  # virtualenv problem pylint: disable=E0611
import logging
import os
import glob
import shutil
import sys
import re

from avocado.utils import cpu
from avocado.utils import distro
from avocado.utils import genio
from avocado.utils import linux_modules
from avocado.utils import path as utils_path
from avocado.utils import process

from . import data_dir
from . import asset
from . import cartesian_config
from . import utils_selinux
from . import defaults
from . import arch
from .compat import get_opt


LOG = logging.getLogger("avocado.app")

basic_program_requirements = ['xz', 'tcpdump', 'nc', 'ip', 'arping', 'diff']

recommended_programs = {'qemu': [('qemu-kvm', 'kvm'), ('qemu-img',),
                                 ('qemu-io',), ('virsh',)],
                        'spice': [('qemu-kvm', 'kvm'), ('qemu-img',),
                                  ('qemu-io',)],
                        'libvirt': [('virsh',), ('virt-install',),
                                    ('fakeroot',), ('semanage',),
                                    ('getfattr',), ('restorecon',)],
                        'openvswitch': [],
                        'lvsb': [('semanage',), ('getfattr',), ('restorecon',), ('virt-sandbox')],
                        'v2v': [],
                        'libguestfs': [('perl',)]}

mandatory_programs = {'qemu': basic_program_requirements + ['gcc'],
                      'spice': basic_program_requirements + ['gcc'],
                      'libvirt': basic_program_requirements,
                      'openvswitch': basic_program_requirements,
                      'lvsb': ['virt-sandbox', 'virt-sandbox-service', 'virsh'],
                      'v2v': basic_program_requirements,
                      'libguestfs': basic_program_requirements}

mandatory_headers = {'qemu': ['Python.h', 'types.h', 'socket.h', 'unistd.h'],
                     'spice': [],
                     'libvirt': [],
                     'openvswitch': [],
                     'v2v': [],
                     'lvsb': [],
                     'libguestfs': []}

first_subtest = {'qemu': ['unattended_install', 'steps'],
                 'spice': ['unattended_install', 'steps'],
                 'libvirt': ['unattended_install'],
                 'openvswitch': ['unattended_install'],
                 'v2v': ['unattended_install'],
                 'libguestfs': ['unattended_install'],
                 'lvsb': []}

last_subtest = {'qemu': ['shutdown'],
                'spice': ['shutdown'],
                'libvirt': ['shutdown', 'remove_guest'],
                'openvswitch': ['shutdown'],
                'v2v': ['shutdown'],
                'libguestfs': ['shutdown'],
                'lvsb': []}

test_filter = ['__init__', 'cfg', 'dropin.py']


def get_guest_os_info_list(test_name, guest_os):
    """
    Returns a list of matching assets compatible with the specified test name
    and guest OS
    """
    os_info_list = []

    cartesian_parser = cartesian_config.Parser()
    cartesian_parser.parse_file(
        data_dir.get_backend_cfg_path(test_name, 'guest-os.cfg'))
    cartesian_parser.only_filter(guest_os)
    dicts = cartesian_parser.get_dicts()

    for params in dicts:
        image_name = params.get('image_name', 'image').split('/')[-1]
        shortname = params.get('shortname', guest_os)
        os_info_list.append({'asset': image_name, 'variant': shortname})

    if not os_info_list:
        LOG.error("Could not find any assets compatible with %s for %s",
                  guest_os, test_name)
        raise ValueError("Missing compatible assets for %s" % guest_os)

    return os_info_list


def get_config_filter():
    config_filter = ['__init__', ]
    for provider_subdir in asset.get_test_provider_subdirs():
        config_filter.append(os.path.join('%s' % provider_subdir, 'cfg'))
    return config_filter


def verify_recommended_programs(t_type):
    cmds = recommended_programs[t_type]
    found = False
    for cmd_aliases in cmds:
        for cmd in cmd_aliases:
            found = None
            try:
                found = utils_path.find_command(cmd)
                LOG.debug('%s OK', found)
                break
            except utils_path.CmdNotFoundError:
                pass
        if not found:
            if len(cmd_aliases) == 1:
                LOG.info("Recommended command %s missing. You may "
                         "want to install it if not building from "
                         "source.", cmd_aliases[0])
            else:
                LOG.info("Recommended command missing. You may "
                         "want to install it if not building it from "
                         "source. Aliases searched: %s", cmd_aliases)


def verify_mandatory_programs(t_type, guest_os):
    failed_cmds = []
    cmds = mandatory_programs[t_type]
    for cmd in cmds:
        try:
            LOG.debug('%s OK', utils_path.find_command(cmd))
        except utils_path.CmdNotFoundError:
            LOG.error("Required command %s is missing. You must "
                      "install it", cmd)
            failed_cmds.append(cmd)

    includes = mandatory_headers[t_type]
    available_includes = glob.glob('/usr/include/*/*')
    for include in available_includes:
        include_basename = os.path.basename(include)
        if include_basename in includes:
            LOG.debug('%s OK', include)
            includes.pop(includes.index(include_basename))

    if includes:
        for include in includes:
            LOG.error("Required include %s is missing. You may have to "
                      "install it", include)

    failures = failed_cmds + includes

    if failures:
        raise ValueError('Missing (cmds/includes): %s' % " ".join(failures))


def write_subtests_files(config_file_list, output_file_object, test_type=None):
    """
    Writes a collection of individual subtests config file to one output file

    Optionally, for tests that we know their type, write the 'virt_test_type'
    configuration automatically.
    """
    if test_type is not None:
        output_file_object.write("    - @type_specific:\n")
        output_file_object.write("        variants subtest:\n")

    for provider_name, config_path in config_file_list:
        config_file = open(config_path, 'r')

        write_test_type_line = False
        write_provider_line = False

        for line in config_file.readlines():
            if line.startswith('- ') and provider_name is not None:
                name, deps = line.split(":")
                name = name[1:].strip()
                if name[0] == "@":
                    name = name[1:]
                line = "- %s.%s:%s" % (provider_name, name, deps)

            # special virt_test_type line output
            if test_type is not None:
                if write_test_type_line:
                    type_line = ("                virt_test_type = %s\n" %
                                 test_type)
                    output_file_object.write(type_line)
                    provider_line = ("                provider = %s\n" %
                                     provider_name)
                    output_file_object.write(provider_line)
                    write_test_type_line = False
                elif line.startswith('- '):
                    write_test_type_line = True
                output_file_object.write("            %s" % line)
            else:
                if write_provider_line:
                    provider_line = ("        provider = %s\n" %
                                     provider_name)
                    output_file_object.write(provider_line)
                    write_provider_line = False
                elif line.startswith('- '):
                    write_provider_line = True
                # regular line output
                output_file_object.write("    %s" % line)

        config_file.close()


def get_directory_structure(rootdir, guest_file, first_variant=None):
    rootdir = rootdir.rstrip(os.sep)
    start = rootdir.rfind(os.sep) + 1
    previous_indent = 0
    indent = 0
    number_variants = 0
    first_variant_offset = 0
    if first_variant:
        guest_file.write("variants:\n")
        guest_file.write("    - %s:\n" % first_variant)
        first_variant_offset = 2
    for path, subdirs, files in os.walk(rootdir):
        folders = path[start:].split(os.sep)
        folders = folders[1:]
        indent = len(folders)
        if indent > previous_indent:
            offset = first_variant_offset + indent + number_variants - 1
            guest_file.write("%svariants:\n" % (4 * offset * " "))
            number_variants += 1
        elif indent < previous_indent:
            number_variants = indent
        indent += number_variants
        try:
            base_folder = folders[-1]
        except IndexError:
            base_folder = []
        base_cfg = "%s.cfg" % base_folder
        base_cfg_path = os.path.join(os.path.dirname(path), base_cfg)
        if os.path.isfile(base_cfg_path):
            base_file = open(base_cfg_path, 'r')
            for line in base_file.readlines():
                offset = first_variant_offset + indent - 1
                guest_file.write("%s%s" % ((4 * offset * " "), line))
        else:
            if base_folder:
                offset = first_variant_offset + indent - 1
                guest_file.write("%s- %s:\n" %
                                 ((4 * offset * " "), base_folder))
        variant_printed = False
        if files:
            files.sort()
            for f in files:
                if f.endswith(".cfg"):
                    bf = f[:len(f) - 4]
                    if bf not in subdirs:
                        if not variant_printed:
                            offset = first_variant_offset + indent
                            guest_file.write("%svariants:\n" %
                                             ((4 * offset * " ")))
                            variant_printed = True
                        base_file = open(os.path.join(path, f), 'r')
                        for line in base_file.readlines():
                            offset = first_variant_offset + indent + 1
                            guest_file.write("%s%s"
                                             % ((4 * offset * " "), line))
        indent -= number_variants
        previous_indent = indent


def sync_download_dir(interactive):
    base_download_dir = data_dir.get_base_download_dir()
    download_dir = data_dir.get_download_dir()
    LOG.debug("Copying downloadable assets file definitions from %s "
              "into %s", base_download_dir, download_dir)
    download_file_list = glob.glob(os.path.join(base_download_dir,
                                                "*.ini"))
    for src_file in download_file_list:
        dst_file = os.path.join(download_dir,
                                os.path.basename(src_file))
        if not os.path.isfile(dst_file):
            shutil.copyfile(src_file, dst_file)
        else:
            diff_cmd = "diff -Naur %s %s" % (dst_file, src_file)
            diff_result = process.run(
                diff_cmd, ignore_status=True, verbose=False)
            if diff_result.exit_status != 0:
                LOG.debug("%s result:\n %s",
                          diff_result.command,
                          diff_result.stdout_text)
                answer = genio.ask('Download file "%s" differs from "%s". '
                                   'Overwrite?' % (dst_file, src_file),
                                   auto=not interactive)
                if answer == "y":
                    LOG.debug("Restoring download file %s from sample",
                              dst_file)
                    shutil.copyfile(src_file, dst_file)
                else:
                    LOG.debug("Preserving existing %s file", dst_file)
            else:
                LOG.debug('Download file %s exists, not touching',
                          dst_file)


def create_guest_os_cfg(t_type):
    guest_os_cfg_dir = os.path.join(data_dir.get_shared_dir(), 'cfg', 'guest-os')
    guest_os_cfg_path = data_dir.get_backend_cfg_path(t_type, 'guest-os.cfg')
    guest_os_cfg_file = open(guest_os_cfg_path, 'w')
    get_directory_structure(guest_os_cfg_dir, guest_os_cfg_file, "Guest")
    LOG.debug("Config file %s auto generated from guest OS samples",
              guest_os_cfg_path)


def host_os_get_distro_name(options, detected):
    """
    Gets the distro name, either from the command line or auto detection

    If option vt_host_distro_compat is set, name is returned as
    uppercase or capitalized

    :param options: parsed command line arguments results
    :type options: :class:`argparse.Namespace`
    :param detected: result of :class:`avocado.utils.distro.detect`
    :type detected: :class:`avocado.utils.distro.LinuxDistro`
    """
    if get_opt(options, 'vt_host_distro_name'):
        return get_opt(options, 'vt_host_distro_name')
    if detected.name == 'rhel':
        return 'RHEL'
    elif detected.name == 'fedora':
        return 'Fedora'
    return "Host_%s" % detected.name


def create_host_os_cfg(options):
    def _forced_or_detected(forced, detected):
        if forced:
            return forced
        else:
            return detected
    host_os_cfg_path = data_dir.get_backend_cfg_path(get_opt(options, 'vt.type'),
                                                     'host.cfg')
    with open(host_os_cfg_path, 'w') as cfg:
        detected = distro.detect()
        name = host_os_get_distro_name(options, detected)
        version = _forced_or_detected(get_opt(options, 'vt_host_distro_version'),
                                      "m%s" % detected.version)
        release = _forced_or_detected(get_opt(options, 'vt_host_distro_release'),
                                      "u%s" % detected.release)
        arch = _forced_or_detected(get_opt(options, 'vt_host_distro_arch'),
                                   "Host_arch_%s" % detected.arch)
        vendor = cpu.get_vendor() if hasattr(cpu, 'get_vendor') else cpu.get_cpu_vendor_name()
        family = None
        if hasattr(cpu, 'get_family'):
            try:
                family = cpu.get_family()
            except Exception:
                pass
        cpu_version = cpu.get_version() if hasattr(cpu, 'get_version') else None

        cfg.write("variants:\n")
        cfg.write("    - @Host:\n")
        cfg.write("        variants:\n")
        cfg.write("            - @%s:\n" % name)
        cfg.write("                variants:\n")
        cfg.write("                    - @%s:\n" % version)
        cfg.write("                        variants:\n")
        cfg.write("                            - @%s:\n" % release)
        cfg.write("                                variants:\n")
        cfg.write("                                    - @%s:\n" % arch)
        if vendor:
            cfg.write("variants:\n")
            cfg.write("    - @HostCpuVendor:\n")
            cfg.write("        variants:\n")
            cfg.write("            - @%s:\n" % vendor)
        if family:
            cfg.write("variants:\n")
            cfg.write("    - @HostCpuFamily:\n")
            cfg.write("        variants:\n")
            cfg.write("            - @%s:\n" % family)
            if cpu_version:
                cfg.write("                variants:\n")
                cfg.write("                    - @HostCpuVersion:\n")
                cfg.write("                        variants:\n")
                cfg.write("                            - @%s:\n" % cpu_version)

    count = [get_opt(options, 'vt_host_distro_name'),
             get_opt(options, 'vt_host_distro_version'),
             get_opt(options, 'vt_host_distro_release'),
             get_opt(options, 'vt_host_distro_arch')].count(None)
    if count == 4:
        source = "distro detection"
    elif count == 0:
        source = "command line parameters"
    else:
        source = "distro detection and command line parameters"
    LOG.debug("Config file %s generated from %s", host_os_cfg_path, source)


def create_subtests_cfg(t_type):
    specific_test_list = []
    specific_file_list = []
    specific_subdirs = asset.get_test_provider_subdirs(t_type)
    provider_names_specific = asset.get_test_provider_names(t_type)
    config_filter = get_config_filter()

    provider_info_specific = []
    for specific_provider in provider_names_specific:
        provider_info_specific.append(
            asset.get_test_provider_info(specific_provider))

    for subdir in specific_subdirs:
        specific_test_list += data_dir.SubdirGlobList(subdir,
                                                      '*.py',
                                                      test_filter)
        specific_file_list += data_dir.SubdirGlobList(subdir,
                                                      '*.cfg',
                                                      config_filter)

    shared_test_list = []
    shared_file_list = []
    shared_subdirs = asset.get_test_provider_subdirs('generic')
    shared_subdirs += asset.get_test_provider_subdirs('multi_host_migration')
    provider_names_shared = asset.get_test_provider_names('generic')
    provider_names_shared += asset.get_test_provider_names('multi_host_migration')

    provider_info_shared = []
    for shared_provider in provider_names_shared:
        provider_info_shared.append(
            asset.get_test_provider_info(shared_provider))

    if not t_type == 'lvsb':
        for subdir in shared_subdirs:
            shared_test_list += data_dir.SubdirGlobList(subdir,
                                                        '*.py',
                                                        test_filter)
            shared_file_list += data_dir.SubdirGlobList(subdir,
                                                        '*.cfg',
                                                        config_filter)

    all_specific_test_list = []
    for test in specific_test_list:
        for p in provider_info_specific:
            provider_base_path = p['backends'][t_type]['path']
            if provider_base_path in test:
                provider_name = p['name']
                break

        basename = os.path.basename(test)
        if basename != "__init__.py":
            all_specific_test_list.append("%s.%s" %
                                          (provider_name,
                                           basename.split(".")[0]))
    all_shared_test_list = []
    for test in shared_test_list:
        for p in provider_info_shared:
            if 'generic' in p['backends']:
                provider_base_path = p['backends']['generic']['path']
                if provider_base_path in test:
                    provider_name = p['name']
                    break
            if 'multi_host_migration' in p['backends']:
                provider_base_path = p['backends']['multi_host_migration']['path']
                if provider_base_path in test:
                    provider_name = p['name']
                    break

        basename = os.path.basename(test)
        if basename != "__init__.py":
            all_shared_test_list.append("%s.%s" %
                                        (provider_name,
                                         basename.split(".")[0]))

    all_specific_test_list.sort()
    all_shared_test_list.sort()

    first_subtest_file = []
    last_subtest_file = []
    non_dropin_tests = []
    tmp = []

    for shared_file in shared_file_list:
        provider_name = None
        for p in provider_info_shared:
            provider_base_path = p['backends']['generic']['path']
            if provider_base_path in shared_file:
                provider_name = p['name']
                break
            provider_base_path = p['backends']['multi_host_migration']['path']
            if provider_base_path in test:
                provider_name = p['name']
                break

        shared_file_obj = open(shared_file, 'r')
        for line in shared_file_obj.readlines():
            line = line.strip()
            if re.match("type\s*=.*", line):
                cartesian_parser = cartesian_config.Parser()
                cartesian_parser.parse_string(line)
                td = next(cartesian_parser.get_dicts())
                values = td['type'].split(" ")
                for value in values:
                    if t_type not in non_dropin_tests:
                        non_dropin_tests.append("%s.%s" %
                                                (provider_name, value))

        shared_file_name = os.path.basename(shared_file)
        shared_file_name = shared_file_name.split(".")[0]
        if shared_file_name in first_subtest[t_type]:
            if [provider_name, shared_file] not in first_subtest_file:
                first_subtest_file.append([provider_name, shared_file])
        elif shared_file_name in last_subtest[t_type]:
            if [provider_name, shared_file] not in last_subtest_file:
                last_subtest_file.append([provider_name, shared_file])
        else:
            if [provider_name, shared_file] not in tmp:
                tmp.append([provider_name, shared_file])
    shared_file_list = tmp

    tmp = []
    for shared_file in specific_file_list:
        provider_name = None
        for p in provider_info_specific:
            provider_base_path = p['backends'][t_type]['path']
            if provider_base_path in shared_file:
                provider_name = p['name']
                break

        shared_file_obj = open(shared_file, 'r')
        for line in shared_file_obj.readlines():
            line = line.strip()
            if re.match("type\s*=.*", line):
                cartesian_parser = cartesian_config.Parser()
                cartesian_parser.parse_string(line)
                td = next(cartesian_parser.get_dicts())
                values = td['type'].split(" ")
                for value in values:
                    if value not in non_dropin_tests:
                        non_dropin_tests.append("%s.%s" %
                                                (provider_name, value))

        shared_file_name = os.path.basename(shared_file)
        shared_file_name = shared_file_name.split(".")[0]
        if shared_file_name in first_subtest[t_type]:
            if [provider_name, shared_file] not in first_subtest_file:
                first_subtest_file.append([provider_name, shared_file])
        elif shared_file_name in last_subtest[t_type]:
            if [provider_name, shared_file] not in last_subtest_file:
                last_subtest_file.append([provider_name, shared_file])
        else:
            if [provider_name, shared_file] not in tmp:
                tmp.append([provider_name, shared_file])
    specific_file_list = tmp

    subtests_cfg = os.path.join(data_dir.get_backend_dir(t_type), 'cfg',
                                'subtests.cfg')
    subtests_file = open(subtests_cfg, 'w')
    subtests_file.write(
        "# Do not edit, auto generated file from subtests config\n")

    subtests_file.write("variants subtest:\n")
    write_subtests_files(first_subtest_file, subtests_file)
    write_subtests_files(specific_file_list, subtests_file, t_type)
    write_subtests_files(shared_file_list, subtests_file)
    write_subtests_files(last_subtest_file, subtests_file)

    subtests_file.close()
    LOG.debug("Config file %s auto generated from subtest samples",
              subtests_cfg)


def create_config_files(test_dir, shared_dir, interactive, t_type, step=None,
                        force_update=False):
    def is_file_tracked(fl):
        tracked_result = process.run("git ls-files %s --error-unmatch" % fl,
                                     ignore_status=True, verbose=False)
        return tracked_result.exit_status == 0

    if step is None:
        step = 0
    LOG.info("")
    step += 1
    LOG.info("%d - Generating config set", step)
    config_file_list = data_dir.SubdirGlobList(os.path.join(test_dir, "cfg"),
                                               "*.cfg",
                                               get_config_filter())
    config_file_list = [cf for cf in config_file_list if is_file_tracked(cf)]
    config_file_list_shared = glob.glob(os.path.join(shared_dir, "cfg",
                                                     "*.cfg"))

    provider_info_specific = []
    provider_names_specific = asset.get_test_provider_names(t_type)
    for specific_provider in provider_names_specific:
        provider_info_specific.append(
            asset.get_test_provider_info(specific_provider))

    specific_subdirs = asset.get_test_provider_subdirs(t_type)
    for subdir in specific_subdirs:
        for p in provider_info_specific:
            if 'cartesian_configs' in p['backends'][t_type]:
                for c in p['backends'][t_type]['cartesian_configs']:
                    cfg = os.path.join(subdir, "cfg", c)
                    config_file_list.append(cfg)

    # Handle overrides of cfg files. Let's say a test provides its own
    # subtest.cfg.sample, this file takes precedence over the shared
    # subtest.cfg.sample. So, yank this file from the cfg file list.

    config_file_list_shared_keep = []
    for cf in config_file_list_shared:
        basename = os.path.basename(cf)
        target = os.path.join(test_dir, "cfg", basename)
        if target not in config_file_list:
            config_file_list_shared_keep.append(cf)

    config_file_list += config_file_list_shared_keep
    for config_file in config_file_list:
        src_file = config_file
        dst_file = os.path.join(test_dir, "cfg", os.path.basename(config_file))
        if not os.path.isfile(dst_file):
            LOG.debug("Creating config file %s from sample", dst_file)
            shutil.copyfile(src_file, dst_file)
        else:
            diff_cmd = "diff -Naur %s %s" % (dst_file, src_file)
            diff_result = process.run(
                diff_cmd, ignore_status=True, verbose=False)
            if diff_result.exit_status != 0:
                LOG.info("%s result:\n %s",
                         diff_result.command,
                         diff_result.stdout_text)
                answer = genio.ask("Config file  %s differs from %s."
                                   "Overwrite?" % (dst_file, src_file),
                                   auto=force_update or not interactive)

                if answer == "y":
                    LOG.debug("Restoring config file %s from sample",
                              dst_file)
                    shutil.copyfile(src_file, dst_file)
                else:
                    LOG.debug("Preserving existing %s file", dst_file)
            else:
                if force_update:
                    update_msg = 'Config file %s exists, equal to sample'
                else:
                    update_msg = 'Config file %s exists, not touching'
                LOG.debug(update_msg, dst_file)
    return step


def haz_defcon(datadir, imagesdir, isosdir, tmpdir):
    """
    Compare current types from Defaults, or if default, compare on-disk type
    """
    # Searching through default contexts is very slow.
    # Exploit restorecon -n to find any defaults
    try:
        # First element is list, third tuple item is desired context
        data_type = utils_selinux.diff_defcon(datadir, False)[0][2]
    except IndexError:  # object matches default, get current on-disk context
        data_type = utils_selinux.get_context_of_file(datadir)
    # Extract just the type component
    data_type = utils_selinux.get_type_from_context(data_type)

    try:
        # Do not descend, we want to know the base-dir def. context
        images_type = utils_selinux.diff_defcon(imagesdir, False)[0][2]
    except IndexError:
        images_type = utils_selinux.get_context_of_file(imagesdir)
    images_type = utils_selinux.get_type_from_context(images_type)

    try:
        isos_type = utils_selinux.diff_defcon(isosdir, False)[0][2]
    except IndexError:
        isos_type = utils_selinux.get_context_of_file(isosdir)
    isos_type = utils_selinux.get_type_from_context(isos_type)

    try:
        tmp_type = utils_selinux.diff_defcon(tmpdir, False)[0][2]
    except IndexError:
        tmp_type = utils_selinux.get_context_of_file(tmpdir)
    tmp_type = utils_selinux.get_type_from_context(tmp_type)

    # hard-coded values b/c only four of them and widly-used
    if data_type == 'virt_var_lib_t':
        if images_type == 'virt_image_t':
            if isos_type == 'virt_content_t':
                if tmp_type == 'user_tmp_t':
                    return True  # No changes needed
    return False


def set_defcon(datadir, imagesdir, isosdir, tmpdir):
    """
    Tries to set datadir default contexts returns True if changed
    """
    made_changes = False
    try:
        # Returns list of tuple(pathname, from, to) of context differences
        # between on-disk and defaults.  Only interested in top-level
        # object [0] and the context it would change to [2]
        data_type = utils_selinux.diff_defcon(datadir, False)[0][2]
        # Extrach only the type
        existing_data = utils_selinux.get_type_from_context(data_type)
    except IndexError:
        existing_data = None
    try:
        images_type = utils_selinux.diff_defcon(imagesdir, False)[0][2]
        existing_images = utils_selinux.get_type_from_context(images_type)
    except IndexError:
        existing_images = None
    try:
        isos_type = utils_selinux.diff_defcon(isosdir, False)[0][2]
        existing_isos = utils_selinux.get_type_from_context(isos_type)
    except IndexError:
        existing_isos = None

    try:
        tmp_type = utils_selinux.diff_defcon(tmpdir, False)[0][2]
        existing_tmp = utils_selinux.get_type_from_context(tmp_type)
    except IndexError:
        existing_tmp = None

    # Only print slow info message one time
    could_be_slow = False
    msg = "Defining default contexts, this could take a few seconds..."
    # Changing default contexts is *slow*, avoid it if not necessary
    if existing_data is None or existing_data != 'virt_var_lib_t':
        # semanage gives errors if don't treat /usr & /usr/local the same
        data_regex = utils_selinux.transmogrify_usr_local(datadir)
        LOG.info(msg)
        could_be_slow = True
        # This applies only to datadir symlink, not sub-directories!
        utils_selinux.set_defcon('virt_var_lib_t', data_regex)
        made_changes = True

    if existing_images is None or existing_images != 'virt_image_t':
        # Applies to imagesdir and everything below
        images_regex = utils_selinux.transmogrify_usr_local(imagesdir)
        images_regex = utils_selinux.transmogrify_sub_dirs(images_regex)
        if not could_be_slow:
            LOG.info(msg)
            could_be_slow = True
        utils_selinux.set_defcon('virt_image_t', images_regex)
        made_changes = True

    if existing_isos is None or existing_isos != 'virt_content_t':
        # Applies to isosdir and everything below
        isos_regex = utils_selinux.transmogrify_usr_local(isosdir)
        isos_regex = utils_selinux.transmogrify_sub_dirs(isos_regex)
        if not could_be_slow:
            LOG.info(msg)
            could_be_slow = True
        utils_selinux.set_defcon('virt_content_t', isos_regex)
        made_changes = True

    if existing_tmp is None or existing_tmp != 'user_tmp_t':
        tmp_regex = utils_selinux.transmogrify_usr_local(tmpdir)
        tmp_regex = utils_selinux.transmogrify_sub_dirs(tmp_regex)
        if not could_be_slow:
            LOG.info(msg)
            could_be_slow = True
        utils_selinux.set_defcon('user_tmp_t', tmp_regex)
        made_changes = True

    return made_changes


def verify_selinux(datadir, imagesdir, isosdir, tmpdir,
                   interactive, selinux=False):
    """
    Verify/Set/Warn about SELinux and default file contexts for testing.

    :param datadir: Abs. path to data-directory symlink
    :param imagesdir: Abs. path to data/images directory
    :param isosdir: Abs. path to data/isos directory
    :param tmpdir: Abs. path to avocado-vt tmp dir
    :param interactive: True if running from console
    :param selinux: Whether setup SELinux contexts for shared/data
    """
    # datadir can be a symlink, but these must not have any
    imagesdir = os.path.realpath(imagesdir)
    isosdir = os.path.realpath(isosdir)
    tmpdir = os.path.realpath(tmpdir)
    needs_relabel = None
    try:
        # Raise SeCmdError if selinux not installed
        if utils_selinux.get_status() == 'enforcing':
            # Check if default contexts are set
            if not haz_defcon(datadir, imagesdir, isosdir, tmpdir):
                if selinux:
                    answer = "y"
                else:
                    answer = genio.ask("Setup all undefined default SE"
                                       "Linux contexts for shared/data/?",
                                       auto=not interactive)
            else:
                answer = "n"
            if answer.lower() == "y":
                # Assume relabeling is needed if changes made
                needs_relabel = set_defcon(datadir, imagesdir, isosdir, tmpdir)
            # Only relabel if files/dirs don't match default
            labels_ok = utils_selinux.verify_defcon(datadir, False)
            labels_ok &= utils_selinux.verify_defcon(imagesdir, True)
            labels_ok &= utils_selinux.verify_defcon(isosdir, True)
            labels_ok &= utils_selinux.verify_defcon(tmpdir, True)
            if labels_ok:
                needs_relabel = False
            else:
                LOG.warning("On-disk SELinux labels do not match defaults")
                needs_relabel = True
        # Disabled or Permissive mode is same result as not installed
        else:
            LOG.info("SELinux in permissive or disabled, testing"
                     "in enforcing mode is highly encourraged.")
    except utils_selinux.SemanageError:
        LOG.info("Could not set default SELinux contexts. Please")
        LOG.info("consider installing the semanage program then ")
        LOG.info("verifying and/or running running:")
        # Paths must be transmogrified (changed) into regular expressions
        LOG.info("semanage fcontext --add -t virt_var_lib_t '%s'",
                 utils_selinux.transmogrify_usr_local(datadir))
        LOG.info("semanage fcontext --add -t virt_image_t '%s'",
                 utils_selinux.transmogrify_usr_local(
                     utils_selinux.transmogrify_sub_dirs(imagesdir)))
        LOG.info("semanage fcontext --add -t virt_content_t '%s'",
                 utils_selinux.transmogrify_usr_local(
                     utils_selinux.transmogrify_sub_dirs(isosdir)))
        LOG.info("semanage fcontext --add -t user_tmp_t '%s'",
                 utils_selinux.transmogrify_usr_local(
                     utils_selinux.transmogrify_sub_dirs(tmpdir)))
        needs_relabel = None  # Next run will catch if relabeling needed
    except utils_selinux.SelinuxError:  # Catchall SELinux related
        LOG.info("SELinux not available, or error in command/setup.")
        LOG.info("Please manually verify default file contexts before")
        LOG.info("testing with SELinux enabled and enforcing.")
    if needs_relabel:
        if selinux:
            answer = "y"
        else:
            answer = genio.ask("Relabel from default contexts?",
                               auto=not interactive)
        if answer.lower() == 'y':
            changes = utils_selinux.apply_defcon(datadir, False)
            changes += utils_selinux.apply_defcon(imagesdir, True)
            changes += utils_selinux.apply_defcon(isosdir, True)
            changes += utils_selinux.apply_defcon(tmpdir, True)
            LOG.info("Corrected contexts on %d files/dirs",
                     len(changes))


def bootstrap(options, interactive=False):
    """
    Common virt test assistant module.

    :param options: Command line options.
    :param interactive: Whether to ask for confirmation.
    """
    if get_opt(options, 'yes_to_all'):
        interactive = False

    vt_type = get_opt(options, 'vt.type')
    LOG.info("Running bootstrap for %s", vt_type)
    step = 0

    LOG.info("")
    step += 1
    LOG.info("%d - Checking the mandatory programs and headers", step)
    guest_os = get_opt(options, 'vt.guest_os') or defaults.DEFAULT_GUEST_OS
    try:
        verify_mandatory_programs(vt_type, guest_os)
    except Exception as details:
        LOG.debug(details)
        LOG.debug('Install the missing programs and/or headers and '
                  're-run boostrap')
        sys.exit(1)

    LOG.info("")
    step += 1
    LOG.info("%d - Checking the recommended programs", step)
    verify_recommended_programs(vt_type)

    LOG.info("")
    step += 1
    LOG.info("%d - Updating test providers repo configuration from local copy", step)
    tp_base_dir = data_dir.get_base_test_providers_dir()
    tp_local_dir = data_dir.get_test_providers_dir()
    dir_util.copy_tree(tp_base_dir, tp_local_dir)

    not_downloaded = asset.test_providers_not_downloaded()
    if not_downloaded:
        action = "Downloading"
    else:
        action = "Updating"
    if not get_opt(options, 'vt_no_downloads'):
        LOG.info("")
        step += 1
        LOG.info("%d - %s the test providers from remote repos", step, action)
        asset.download_all_test_providers(get_opt(options, 'vt_update_providers'))
    else:
        if not_downloaded:
            LOG.warn("The following test providers have not been downloaded: %s",
                     ", ".join(not_downloaded))

    LOG.info("")
    step += 1
    LOG.info("%d - Verifying directories", step)
    datadir = data_dir.get_data_dir()
    shared_dir = data_dir.get_shared_dir()
    sub_dir_list = ["images", "isos", "steps_data", "gpg", "downloads"]
    for sub_dir in sub_dir_list:
        sub_dir_path = os.path.join(datadir, sub_dir)
        if not os.path.isdir(sub_dir_path):
            LOG.debug("Creating %s", sub_dir_path)
            os.makedirs(sub_dir_path)
        else:
            LOG.debug("Dir %s exists, not creating",
                      sub_dir_path)

    base_backend_dir = data_dir.get_base_backend_dir()
    local_backend_dir = data_dir.get_local_backend_dir()
    LOG.info("")
    step += 1
    LOG.info("%d - Syncing backend dirs %s -> %s", step, base_backend_dir,
             local_backend_dir)
    dir_util.copy_tree(base_backend_dir, local_backend_dir)

    sync_download_dir(interactive)

    test_dir = data_dir.get_backend_dir(vt_type)
    if vt_type == 'libvirt':
        step = create_config_files(test_dir, shared_dir, interactive,
                                   vt_type, step,
                                   force_update=get_opt(options, 'vt_update_config'))
        create_subtests_cfg(vt_type)
        create_guest_os_cfg(vt_type)
        # Don't bother checking if changes can't be made
        if os.getuid() == 0:
            verify_selinux(datadir,
                           os.path.join(datadir, 'images'),
                           os.path.join(datadir, 'isos'),
                           data_dir.get_tmp_dir(),
                           interactive, get_opt(options, 'vt_selinux_setup'))

    # lvsb test doesn't use any shared configs
    elif vt_type == 'lvsb':
        create_subtests_cfg(vt_type)
        if os.getuid() == 0:
            # Don't bother checking if changes can't be made
            verify_selinux(datadir,
                           os.path.join(datadir, 'images'),
                           os.path.join(datadir, 'isos'),
                           data_dir.get_tmp_dir(),
                           interactive, get_opt(options, 'vt_selinux_setup'))
    else:  # Some other test
        step = create_config_files(test_dir, shared_dir, interactive,
                                   vt_type, step,
                                   force_update=get_opt(options, 'vt_update_config'))
        create_subtests_cfg(vt_type)
        create_guest_os_cfg(vt_type)
    create_host_os_cfg(options)

    if not (get_opt(options, 'vt_no_downloads') or
            get_opt(options, 'vt_skip_verify_download_assets')):
        LOG.info("")
        step += 1
        LOG.info("%s - Verifying (and possibly downloading) guest image",
                 step)
        try:
            for os_info in get_guest_os_info_list(vt_type, guest_os):
                os_asset = os_info['asset']
                try:
                    asset.download_asset(os_asset, interactive=interactive,
                                         restore_image=True)
                except AssertionError:
                    pass    # Not all files are managed via asset

        except ValueError as details:
            LOG.error(details)
            sys.exit(1)

    check_modules = []
    if vt_type == "qemu":
        check_modules = arch.get_kvm_module_list()
    elif vt_type == "openvswitch":
        check_modules = ["openvswitch"]

    if check_modules:
        LOG.info("")
        step += 1
        LOG.info("%d - Checking for modules %s", step,
                 ", ".join(check_modules))
        for module in check_modules:
            if not linux_modules.module_is_loaded(module):
                LOG.warning("Module %s is not loaded. You might want to "
                            "load it", module)
            else:
                LOG.debug("Module %s loaded", module)

    LOG.info("")
    LOG.info("VT-BOOTSTRAP FINISHED")
    LOG.debug("You may take a look at the following online docs for more info:")
    LOG.debug(" - http://avocado-vt.readthedocs.org/")
    LOG.debug(" - http://avocado-framework.readthedocs.org/")



"""
Module simplifying manipulation of XML described at
http://libvirt.org/formatbackup.html
"""
from virttest import xml_utils

from virttest.libvirt_xml import base, accessors, xcepts


class BackupXML(base.LibvirtXMLBase):
    """
    Domain backup XML class

    Properties:
        mode:
            string (push/pull), backup mode
        incremental:
            string, incremental backup based on which checkpoint
        server:
            dict, keys: transport, socket, name, port
    """

    __slots__ = ('mode', 'incremental', 'server', 'disks')

    __uncompareable__ = base.LibvirtXMLBase.__uncompareable__

    __schema_name__ = "domainbackup"

    def __init__(self, virsh_instance=base.virsh):
        accessors.XMLAttribute('mode', self, parent_xpath='/',
                               tag_name='domainbackup', attribute='mode')
        accessors.XMLElementText('incremental', self, parent_xpath='/',
                                 tag_name='incremental')
        accessors.XMLElementDict('server', self, parent_xpath='/',
                                 tag_name='server')
        super(self.__class__, self).__init__(virsh_instance=virsh_instance)
        self.xml = '<domainbackup/>'

    def set_disks(self, value_list):
        """
        Set the disk sub-elements to the backup xml

        :param value_list: The list of the disk info
        """
        for value in value_list:
            value_type = type(value)
            if not issubclass(value_type, self.DiskXML):
                raise xcepts.LibvirtXMLError("Value '%s' Must be a (sub)class of DiskXML,"
                                             "but not a '%s' type"
                                             % (str(value), str(value_type)))
        exist_disks = self.xmltreefile.find('disks')
        if exist_disks is not None:
            self.del_disks()
        if len(value_list) > 0:
            disks_element = xml_utils.ElementTree.SubElement(
                    self.xmltreefile.getroot(), 'disks')
            for disk in value_list:
                disk_element = disk.xmltreefile.getroot()
                disks_element.append(disk_element)
        self.xmltreefile.write()

    def del_disks(self):
        """
        Del all disk sub-elements form the backup xml
        """
        self.xmltreefile.remove_by_xpath('/disks', remove_all=True)
        self.xmltreefile.write()

    class DiskXML(base.LibvirtXMLBase):
        """
        Disk XML class, used to set disk specific parameters when do backup

        Properties:
            name:
                string, disk name
            backup:
                string (yes/no), whether do backup for this disk
            exportname:
                string, export name of NBD, only valid for pull-mode backup
            exportbitmap:
                string, export bitmap name, only valid for pull-mode backup
            type:
                string (file/block/network), describe the type of the disk,
                'network' is not vaild for now due to bz:
                https://bugzilla.redhat.com/show_bug.cgi?id=1812100
            backupmode:
                string, indicate if the disk will do full or incremental backup
            incremental:
                string, indicate the incremental backup is from which checkpoint
            target:
                dict, keys: dev, file. The backup destination
            driver:
                dict, keys: type. Value of type could be qcow2/raw or
                other format
            scratch:
                dict, keys: file, dev. The scratch file/dev used when do
                pull-mode backup
        """

        __slots__ = ('name', 'backup', 'exportname', 'exportbitmap',
                     'type', 'backupmode', 'incremental', 'target',
                     'driver', 'scratch')

        def __init__(self, virsh_instance=base.virsh):
            accessors.XMLAttribute('name', self, parent_xpath='/',
                                   tag_name='disk', attribute='name')
            accessors.XMLAttribute('backup', self, parent_xpath='/',
                                   tag_name='disk', attribute='backup')
            accessors.XMLAttribute('exportname', self, parent_xpath='/',
                                   tag_name='disk', attribute='exportname')
            accessors.XMLAttribute('exportbitmap', self, parent_xpath='/',
                                   tag_name='disk', attribute='exportbitmap')
            accessors.XMLAttribute('type', self, parent_xpath='/',
                                   tag_name='disk', attribute='type')
            accessors.XMLAttribute('backupmode', self, parent_xpath='/',
                                   tag_name='disk', attribute='backupmode')
            accessors.XMLAttribute('incremental', self, parent_xpath='/',
                                   tag_name='disk', attribute='incremental')
            accessors.XMLElementDict('target', self, parent_xpath='/',
                                     tag_name='target')
            accessors.XMLElementDict('driver', self, parent_xpath='/',
                                     tag_name='driver')
            accessors.XMLElementDict('scratch', self, parent_xpath='/',
                                     tag_name='scratch')
            super(self.__class__, self).__init__(virsh_instance=virsh_instance)
            self.xml = "<disk/>"



"""
Utility functions to deal with ppm (qemu screendump format) files.

:copyright: Red Hat 2008-2009
"""

from __future__ import division
import os
import struct
import time
import re
import glob
import logging

from functools import reduce


try:
    from PIL import Image
    from PIL import ImageOps
    from PIL import ImageDraw
    from PIL import ImageFont
except ImportError:
    Image = None
    logging.warning('No python imaging library installed. Screendump '
                    'and Windows guest BSOD detection are disabled. '
                    'In order to enable it, please install python-imaging or '
                    'the equivalent for your distro.')
# Prevent logs pollution
if Image is not None:
    for _logger_name in logging.root.manager.loggerDict:
        if _logger_name.split('.', 1)[0] == 'PIL':
            _pil_logger = logging.getLogger(_logger_name)
            _pil_logger.setLevel(logging.CRITICAL)


try:
    import hashlib
except ImportError:
    import md5

# Some directory/filename utils, for consistency


def _md5eval(data):
    """
    Returns a md5 hash evaluator. This function is implemented in order to
    encapsulate objects in a way that is compatible with python 2.4 and
    python 2.6 without warnings.

    :param data: Optional input string that will be used to update the object.
    """
    try:
        hsh = hashlib.new('md5')
    except NameError:
        hsh = md5.new()
    if data:
        hsh.update(data)

    return hsh


def find_id_for_screendump(md5sum, data_dir):
    """
    Search dir for a PPM file whose name ends with md5sum.

    :param md5sum: md5 sum string
    :param dir: Directory that holds the PPM files.
    :return: The file's basename without any preceding path, e.g.
             ``20080101_120000_d41d8cd98f00b204e9800998ecf8427e.ppm``
    """
    try:
        files = os.listdir(data_dir)
    except OSError:
        files = []
    for fl in files:
        exp = re.compile(r"(.*_)?" + md5sum + r"\.ppm", re.IGNORECASE)
        if exp.match(fl):
            return fl


def generate_id_for_screendump(md5sum, data_dir):
    """
    Generate a unique filename using the given MD5 sum.

    :return: Only the file basename, without any preceding path. The
             filename consists of the current date and time, the MD5 sum and a
             ``.ppm`` extension, e.g.
             ``20080101_120000_d41d8cd98f00b204e9800998ecf8427e.ppm``.
    """
    filename = time.strftime("%Y%m%d_%H%M%S") + "_" + md5sum + ".ppm"
    return filename


def get_data_dir(steps_filename):
    """
    Return the data dir of the given steps filename.
    """
    filename = os.path.basename(steps_filename)
    return os.path.join(os.path.dirname(steps_filename), "..", "steps_data",
                        filename + "_data")


# Functions for working with PPM files

def image_read_from_ppm_file(filename):
    """
    Read a PPM image.

    :return: A 3 element tuple containing the width, height and data of the
            image.
    """
    with open(filename, "rb") as fin:
        fin.readline()
        l2 = fin.readline()
        fin.readline()
        data = fin.read()

    (w, h) = list(map(int, l2.split()))
    return (w, h, data)


def image_write_to_ppm_file(filename, width, height, data):
    """
    Write a PPM image with the given width, height and data.

    :param filename: PPM file path
    :param width: PPM file width (pixels)
    :param height: PPM file height (pixels)
    """
    with open(filename, "wb") as fout:
        fout.write(b"P6\n")
        fout.write(("%d %d\n" % (width, height)).encode())
        fout.write(b"255\n")
        fout.write(data)


def image_crop(width, height, data, x1, y1, dx, dy):
    """
    Crop an image.

    :param width: Original image width
    :param height: Original image height
    :param data: Image data
    :param x1: Desired x coordinate of the cropped region
    :param y1: Desired y coordinate of the cropped region
    :param dx: Desired width of the cropped region
    :param dy: Desired height of the cropped region
    :return: A 3-tuple containing the width, height and data of the
             cropped image.
    """
    if x1 > width - 1:
        x1 = width - 1
    if y1 > height - 1:
        y1 = height - 1
    if dx > width - x1:
        dx = width - x1
    if dy > height - y1:
        dy = height - y1
    newdata = b""
    index = (x1 + y1 * width) * 3
    for _ in range(dy):
        newdata += data[index:(index + dx * 3)]
        index += width * 3
    return (dx, dy, newdata)


def image_md5sum(width, height, data):
    """
    Return the md5sum of an image.

    :param width: PPM file width
    :param height: PPM file height
    :param data: PPM file data
    """
    header = "P6\n%d %d\n255\n" % (width, height)
    hsh = _md5eval(header.encode())
    hsh.update(data)
    return hsh.hexdigest()


def get_region_md5sum(width, height, data, x1, y1, dx, dy,
                      cropped_image_filename=None):
    """
    Return the md5sum of a cropped region.

    :param width: Original image width
    :param height: Original image height
    :param data: Image data
    :param x1: Desired x coord of the cropped region
    :param y1: Desired y coord of the cropped region
    :param dx: Desired width of the cropped region
    :param dy: Desired height of the cropped region
    :param cropped_image_filename: if not None, write the resulting cropped
            image to a file with this name
    """
    (cw, ch, cdata) = image_crop(width, height, data, x1, y1, dx, dy)
    # Write cropped image for debugging
    if cropped_image_filename:
        image_write_to_ppm_file(cropped_image_filename, cw, ch, cdata)
    return image_md5sum(cw, ch, cdata)


def image_verify_ppm_file(filename):
    """
    Verify the validity of a PPM file.

    :param filename: Path of the file being verified.
    :return: True if filename is a valid PPM image file. This function
             reads only the first few bytes of the file so it should be rather
             fast.
    """
    try:
        size = os.path.getsize(filename)
        with open(filename, "rb") as fin:
            assert(fin.readline().strip() == b"P6")
            (width, height) = map(int, fin.readline().split())
            assert(width > 0 and height > 0)
            assert(fin.readline().strip() == b"255")
            size_read = fin.tell()
        assert(size - size_read == width * height * 3)
        return True
    except Exception:
        return False


def image_comparison(width, height, data1, data2):
    """
    Generate a green-red comparison image from two given images.

    :param width: Width of both images
    :param height: Height of both images
    :param data1: Data of first image
    :param data2: Data of second image
    :return: A 3-element tuple containing the width, height and data of the
            generated comparison image.

    :note: Input images must be the same size.
    """
    newdata = ""
    i = 0
    while i < width * height * 3:
        # Compute monochromatic value of current pixel in data1
        pixel1_str = data1[i:i + 3]
        temp = struct.unpack("BBB", pixel1_str)
        value1 = int((temp[0] + temp[1] + temp[2]) / 3)
        # Compute monochromatic value of current pixel in data2
        pixel2_str = data2[i:i + 3]
        temp = struct.unpack("BBB", pixel2_str)
        value2 = int((temp[0] + temp[1] + temp[2]) / 3)
        # Compute average of the two values
        value = int((value1 + value2) / 2)
        # Scale value to the upper half of the range [0, 255]
        value = 128 + value // 2
        # Compare pixels
        if pixel1_str == pixel2_str:
            # Equal -- give the pixel a greenish hue
            newpixel = [0, value, 0]
        else:
            # Not equal -- give the pixel a reddish hue
            newpixel = [value, 0, 0]
        newdata += struct.pack("BBB", newpixel[0],
                               newpixel[1], newpixel[2])
        i += 3
    return (width, height, newdata)


def image_fuzzy_compare(width, height, data1, data2):
    """
    Return the degree of equality of two given images.

    :param width: Width of both images
    :param height: Height of both images
    :param data1: Data of first image
    :param data2: Data of second image
    :return: Ratio equal_pixel_count / total_pixel_count.

    :note: Input images must be the same size.
    """
    equal = 0.0
    different = 0.0
    i = 0
    while i < width * height * 3:
        pixel1_str = data1[i:i + 3]
        pixel2_str = data2[i:i + 3]
        # Compare pixels
        if pixel1_str == pixel2_str:
            equal += 1.0
        else:
            different += 1.0
        i += 3
    return equal / (equal + different)


def image_average_hash(image, img_wd=8, img_ht=8):
    """
    Resize and convert the image, then get image data as sequence object,
    calculate the average hash
    :param image: an image path or an opened image object
    """
    if not isinstance(image, Image.Image):
        image = Image.open(image)
    image = image.resize((img_wd, img_ht), Image.ANTIALIAS).convert('L')
    avg = reduce(lambda x, y: x + y, image.getdata()) / (img_wd * img_ht)

    def _hta(i):
        if i < avg:
            return 0
        else:
            return 1
    return reduce(lambda x, y_z: x | (y_z[1] << y_z[0]),
                  enumerate(map(_hta, image.getdata())), 0)


def cal_hamming_distance(h1, h2):
    """
    Calculate the hamming distance
    """
    h_distance, distance = 0, h1 ^ h2
    while distance:
        h_distance += 1
        distance &= distance - 1
    return h_distance


def img_ham_distance(base_img, comp_img):
    """
    Calculate two images hamming distance
    """
    base_img_ahash = image_average_hash(base_img)
    comp_img_ahash = image_average_hash(comp_img)
    return cal_hamming_distance(comp_img_ahash, base_img_ahash)


def img_similar(base_img, comp_img, threshold=10):
    """
    check whether two images are similar by hamming distance
    """
    try:
        hamming_distance = img_ham_distance(base_img, comp_img)
    except IOError:
        return False

    if hamming_distance < threshold:
        return True
    else:
        return False


def have_similar_img(base_img, comp_img_path, threshold=10):
    """
    Check whether comp_img_path have a image looks like base_img.
    """
    support_img_format = ['jpg', 'jpeg', 'gif', 'png', 'pmp']
    comp_images = []
    if os.path.isdir(comp_img_path):
        for ext in support_img_format:
            comp_images.extend([os.path.join(comp_img_path, x) for x in
                                glob.glob1(comp_img_path, '*.%s' % ext)])
    else:
        comp_images.append(comp_img_path)

    for img in comp_images:
        if img_similar(base_img, img, threshold):
            return True
    return False


def image_crop_save(image, new_image, box=None):
    """
    Crop an image and save it to a new image.

    :param image: Full path of the original image
    :param new_image: Full path of the cropped image
    :param box: A 4-tuple defining the left, upper, right, and lower pixel coordinate.
    :return: True if crop and save image succeed
    """
    img = Image.open(image)
    if not box:
        x, y = img.size
        box = (x/4, y/4, x*3/4, y*3/4)
    try:
        img.crop(box).save(new_image)
    except (KeyError, SystemError) as e:
        logging.error("Fail to crop image: %s", e)
        return False
    return True


def image_histogram_compare(image_a, image_b, size=(0, 0)):
    """
    Compare the histogram of two images and return similar degree.

    :param image_a: Full path of the first image
    :param image_b: Full path of the second image
    :param size: Convert image to size(width, height), and if size=(0, 0), the function will convert the big size image align with the small one.
    """
    img_a = Image.open(image_a)
    img_b = Image.open(image_b)
    if not any(size):
        size = tuple(map(max, img_a.size, img_b.size))
    img_a_h = img_a.resize(size).convert('RGB').histogram()
    img_b_h = img_b.resize(size).convert('RGB').histogram()
    s = 0
    for i, j in list(zip(img_a_h, img_b_h)):
        if i == j:
            s += 1
        else:
            s += 1 - float(abs(i - j))/max(i, j)
    return s / len(img_a_h)


def add_timestamp(image, timestamp, margin=2):
    """
    Return an image object with timestamp bar added at the bottom.

    param image: pillow image object
    param timestamp: timestamp in seconds since the Epoch
    param margin: timestamp margin, default is 2
    """
    width, height = image.size
    font = ImageFont.load_default()
    watermark = time.strftime('%c', time.localtime(timestamp))
    # bar height = text height + top margin + bottom margin
    bar_height = font.getsize(watermark)[1] + 2 * margin

    # place bar at the bottom
    new_image = ImageOps.expand(image, border=(0, 0, 0, bar_height),
                                fill='lightgrey')
    draw = ImageDraw.Draw(new_image)
    # place timestamp at the left side of the bar
    x, y = margin, height + margin
    draw.text((x, y), watermark, font=font, fill='black')
    return new_image


def image_size(image):
    """
    Return image's size as a tuple (width, height).

    :param image: image file.
    """

    img = Image.open(image)
    return img.size



#!/usr/bin/python
#
# Copyright 2010 Red Hat, Inc. and/or its affiliates.
#
# Licensed to you under the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.  See the files README and
# LICENSE_GPL_v2 which accompany this distribution.
#

import os
import struct
import platform
import optparse

try:
    import hashlib
except ImportError:
    import md5


class ShakeHandError(Exception):

    def __init__(self, msg):
        Exception.__init__(self, msg)
        self.msg = msg

    def __str__(self):
        return ("Shake hand fail. %s" % self.msg)


class Md5MissMatch(Exception):

    def __init__(self, md5_pre, md5_post):
        Exception.__init__(self, md5_pre, md5_post)
        self.md5_pre = md5_pre
        self.md5_post = md5_post

    def __str__(self):
        return ("Md5 miss match. Original md5 = %s, current md5 = %s" %
                (self.md5_pre, self.md5_post))


class VirtIoChannel:

    # Python on Windows 7 return 'Microsoft' rather than 'Windows' as documented.
    is_windows = (platform.system() == 'Windows') or (platform.system() == 'Microsoft')

    def __init__(self, device_name):
        self.ack_format = "3s"
        self.ack_msg = b"ACK"
        self.hi_format = "2s"
        self.hi_msg = b"HI"
        self.recv_msg = b'ALLRECEIVED'
        if self.is_windows:
            vport_name = '\\\\.\\Global\\' + device_name
            from windows_support import WinBufferedReadFile
            self._vport = WinBufferedReadFile(vport_name)
        else:
            vport_name = '/dev/virtio-ports/' + device_name
            self._vport = os.open(vport_name, os.O_RDWR)

    def close(self):
        if self.is_windows:
            self._vport.close()
        else:
            os.close(self._vport)

    def receive(self, size):
        if self.is_windows:
            txt = self._vport.read(size)
        else:
            txt = os.read(self._vport, size)
        return txt

    def send(self, message):
        if self.is_windows:
            self._vport.write(message)
        else:
            os.write(self._vport, message)

    def shake_hand(self, size=0, action="receive"):
        hi_msg_len = struct.calcsize(self.hi_format)
        ack_msg_len = struct.calcsize(self.ack_format)
        if action == "send":
            self.send(self.hi_msg)
            txt = self.receive(hi_msg_len)
            out = struct.unpack(self.hi_format, txt)[0]
            if out != b"HI":
                raise ShakeHandError("Fail to get HI from host.")
            size_s = struct.pack("q", size)
            self.send(size_s)
            txt = self.receive(ack_msg_len)
            ack_str = struct.unpack(self.ack_format, txt)[0]
            if ack_str != self.ack_msg:
                raise ShakeHandError("Host didn't ACK the file size message.")
            return size
        elif action == "receive":
            txt = b''
            while len(txt) < hi_msg_len:
                txt += self.receive(hi_msg_len)
            hi_str = struct.unpack(self.hi_format, txt)[0]
            if hi_str != self.hi_msg:
                raise ShakeHandError("Fail to get HI from host.")
            self.send(txt)
            size = self.receive(8)
            if size:
                size = struct.unpack("q", size)[0]
                txt = struct.pack(self.ack_format, self.ack_msg)
                self.send(txt)
            return size


def md5_init(data=None):
    """
    Returns md5. This function is implemented in order to encapsulate hash
    objects in a way that is compatible with python 2.4 and python 2.6
    without warnings.

    Note that even though python 2.6 hashlib supports hash types other than
    md5 and sha1, we are artificially limiting the input values in order to
    make the function to behave exactly the same among both python
    implementations.

    :param data: Optional input string that will be used to update the hash.
    """

    try:
        md5_value = hashlib.new("md5")
    except NameError:
        md5_value = md5.new()
    if data:
        md5_value.update(data)
    return md5_value


def get_md5(filename, size=None):
    """
    Calculate the hash of filename.
    If size is not None, limit to first size bytes.
    Throw exception if something is wrong with filename.
    Can be also implemented with bash one-liner (assuming size%1024==0):
    dd if=filename bs=1024 count=size/1024 | sha1sum -

    :param filename: Path of the file that will have its hash calculated.
    :param method: Method used to calculate the hash. Supported methods:
            * md5
            * sha1
    :return: Hash of the file, if something goes wrong, return None.
    """
    chunksize = 4096
    fsize = os.path.getsize(filename)

    if not size or size > fsize:
        size = fsize
    f = open(filename, 'rb')

    md5_value = md5_init()
    while size > 0:
        if chunksize > size:
            chunksize = size
        data = f.read(chunksize)
        if len(data) == 0:
            print("Nothing left to read but size=%d" % size)
            break
        md5_value.update(data)
        size -= len(data)
    f.close()
    return md5_value.hexdigest()


def receive(device, filename, p_size=1024):
    recv_size = 0
    vio = VirtIoChannel(device)
    size = vio.shake_hand(action="receive")
    if p_size < int(size):
        p_size = int(size)
    md5_value = md5_init()
    file_no = open(filename, 'wb')
    try:
        while recv_size < size:
            txt = vio.receive(p_size)
            md5_value.update(txt)
            file_no.write(txt)
            recv_size += len(txt)
        vio.send(vio.recv_msg)
    finally:
        file_no.close()
        if vio:
            vio.close()
    md5_sum = md5_value.hexdigest()
    return md5_sum


def send(device, filename, p_size=1024):
    send_size = 0
    f_size = os.path.getsize(filename)
    vio = VirtIoChannel(device)
    vio.shake_hand(f_size, action="send")
    md5_value = md5_init()
    file_no = open(filename, 'rb')
    try:
        while send_size < f_size:
            txt = file_no.read(p_size)
            vio.send(txt)
            md5_value.update(txt)
            send_size += len(txt)
    finally:
        print("Sent size = %s" % send_size)
        file_no.close()
        if vio:
            vio.close()
    md5_sum = md5_value.hexdigest()
    return md5_sum


if __name__ == "__main__":
    txt = "Transfer data between guest and host through virtio serial."
    parser = optparse.OptionParser(txt)

    parser.add_option("-d", "--device", dest="device",
                      help="serial device name used in qemu command"
                           "eg: -device virtserialport,chardev=x,name=redhat"
                           "need use redhat here.")
    parser.add_option("-f", "--filename", dest="filename",
                      help="File transfer to guest or save data to.")
    parser.add_option("-a", "--action", dest="action", default="send",
                      help="Send data out or receive data.")
    parser.add_option("-p", "--package", dest="package", default=1024,
                      help="Package size during file transfer.")

    options, args = parser.parse_args()

    if options.device:
        device = options.device
    else:
        parser.error("Please set -d parameter.")

    if options.filename:
        filename = options.filename
    else:
        parser.error("Please set -f parameter.")
    p_size = options.package
    action = options.action
    if action not in ("receive", "send", "both"):
        parser.error('Please set -a parameter: "receive", "send", "both"')

    if action == "receive":
        md5_sum = receive(device, filename, p_size=p_size)
        print("md5_sum = %s" % md5_sum)
    elif action == "send":
        md5_sum = send(device, filename, p_size=p_size)
        print("md5_sum = %s" % md5_sum)
    else:
        md5_ori = receive(device, filename, p_size=p_size)
        print("md5_original = %s" % md5_ori)
        md5_post = send(device, filename, p_size=p_size)
        print("md5_post = %s" % md5_post)
        if md5_ori != md5_post:
            raise Md5MissMatch(md5_ori, md5_post)
